describe QueryDam do
  describe 'simple query' do
    let!(:relation) { Item.where(name: 'pencil') }
    let!(:key) { QueryDam.watch_relation(relation) }

    specify 'item enters scope on creation' do
      item = Item.create(name: 'pencil')

      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [item], exclusions: []})

      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [], exclusions: []})
    end

    specify 'item enters scope on update' do
      item = Item.create

      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [], exclusions: []})

      item.update_attributes!(name: 'pencil')

      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [item], exclusions: []})
    end

    specify 'item leaves scope on update' do
      item = Item.create(name: 'pencil')
      QueryDam.get_updates(key)

      item.update_attributes!(name: 'pen')
      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [], exclusions: [item.id]})
    end

    specify 'item leaves scope on deletion' do
      item = Item.create(name: 'pencil')
      QueryDam.get_updates(key)

      item.destroy
      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [], exclusions: [item.id]})
    end

    specify 'items gets updated' do
      item = Item.create(name: 'pencil')
      QueryDam.get_updates(key)

      item.update_attributes!(quantity: 1)
      expect(QueryDam.get_updates(key)).to eq(
        {model: Item, updates: [item], exclusions: []})
    end
  end

  describe 'query with limit/offset' do
    let!(:item_a) { Item.create!(name: 'a') }
    let!(:item_b) { Item.create!(name: 'b') }
    let!(:item_c) { Item.create!(name: 'c') }
    let!(:item_d) { Item.create!(name: 'd') }

    let!(:relation) { Item.order('name asc').limit(2).offset(2) }
    let!(:key) { QueryDam.watch_relation(relation) }

    specify 'item is not on page, but shifts it' do
      Item.create!(name: 'a1')

      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [item_b], exclusions: [item_d.id]})
    end

    specify 'item appears on page' do
      item_b1 = Item.create(name: 'b1')
      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [item_b1], exclusions: [item_d.id]})
    end

    specify 'item on page is updated' do
      item_c.update_attributes(quantity: 1)
      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [item_c], exclusions: []})
    end

    specify 'two updates lead to no changes' do
      item_c1 = Item.create(name: 'c1')
      item_c1.destroy
      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [], exclusions: []})
    end

    specify 'item is updated and removed' do
      item_d.update_attributes!(quantity: 1)
      item_d.destroy
      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [], exclusions: [item_d.id]})
    end

    specify 'item is added, updated and removed' do
      item_c1 = Item.create(name: 'c1')
      item_c1.update_attributes(quantity: 1)
      item_c1.destroy
      expect(QueryDam.get_updates(key)).to eq (
        {model: Item, updates: [], exclusions: []})
    end
  end

  describe 'expiration' do
    before :all do
      QueryDam.expiration = 2
    end

    let!(:relation) { Item.where(name: 'pen') }
    let!(:key) { QueryDam.watch_relation(relation) }

    it 'returns nil if expiration time since creation passed' do
      sleep QueryDam.expiration
      expect(QueryDam.get_updates(key)).to be_nil

      expect(Redis::HashKey.new(key).exists?).to be_falsey
      QueryDam.query_sets(key).each do |set|
        expect(set.exists?).to be_falsey
      end
    end

    it 'returns nil if expiration time since last request passed' do
      expect(QueryDam.get_updates(key)).to be_present
      sleep QueryDam.expiration
      expect(QueryDam.get_updates(key)).to be_nil

      expect(Redis::HashKey.new(key).exists?).to be_falsey
      QueryDam.query_sets(key).each do |set|
        expect(set.exists?).to be_falsey
      end
    end
  end
end
